import xml.etree.ElementTree as ET
from sqlalchemy import select, insert, update, delete

from pyjabber.db.database import DB
from pyjabber import metadata
from pyjabber.db.model import Model
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class Roster(metaclass=Singleton):
    """
    Roster plugin.

    Manages the roster list for each registered user in the server.

    A roster in XMPP is a server-stored contact list that manages contacts, presence status,
    and subscription requests.
    It enables real-time presence updates, contact organization, and synchronization across devices,
    ensuring seamless and private communication.
    """

    def __init__(self) -> None:
        self._handlers = {
            "get": self.handle_get,
            "set": self.handle_set,
            "result": self.handle_result
        }
        self._ns = {
            "ns": "jabber:iq:roster",
            "query": "{jabber:iq:roster}query",
            "item": "{jabber:iq:roster}item"
        }

        self._roster_in_memory = {}
        self._update_roster()

    def create_roster_entry(self, jid: JID,  to: JID):
        iq = IQ(from_=str(jid), type_=IQ.TYPE.SET)
        query = ET.SubElement(iq, "{jabber:iq:roster}query")
        if to.domain == metadata.HOST:
            ET.SubElement(query, "{jabber:iq:roster}item", attrib={"jid": to.user, "subscription": "none"})
        else:
            ET.SubElement(query, "{jabber:iq:roster}item", attrib={"jid": to.bare(), "subscription": "none"})

        return self.feed(jid, iq)

    # def check_pending_sub_to(self, jid: JID, to: str) -> ET.Element:
    #     with DB.connection() as con:
    #         res = con.execute("SELECT * FROM pendingsub WHERE jid = ?", (jid,))
    #         res = res.fetchone()
    #     if res:
    #         return res

    @staticmethod
    def store_pending_sub(to_: str, item: ET.Element) -> None:
        with DB.connection() as con:
            query = insert(Model.PendingSubs).values({
                "jid": to_,
                "item":  ET.tostring(item).decode()
            })
            con.execute(query)
            # con.execute("INSERT INTO pendingsub values (?, ?)", (to_, ET.tostring(item).decode()))
            con.commit()

    def update_item(self, item: ET.Element, id_: int):
        with DB.connection() as con:
            query = update(Model.Roster).where(Model.Roster.c.id == id_).values({
                "roster_item": ET.tostring(item).decode()
            })
            con.execute(query)
            # con.execute("UPDATE roster SET rosterItem = ? WHERE id = ?",
            #             (ET.tostring(item).decode(), id_))
            con.commit()
        self._update_roster()

    def _update_roster(self):
        with DB.connection() as con:
            query = select(Model.Roster.c.id, Model.Roster.c.jid, Model.Roster.c.roster_item)
            res = con.execute(query).fetchall()
            # res = con.execute("SELECT id, jid, rosterItem FROM roster", ())
            # res = res.fetchall()
        self._roster_in_memory.clear()
        for id_, jid, item in res:
            if jid not in self._roster_in_memory:
                self._roster_in_memory[jid] = []
            self._roster_in_memory[jid].append({"id": id_, "item": item})

    def roster_by_jid(self, jid: JID):
        if jid.domain == metadata.HOST:
            return self._roster_in_memory.get(jid.user) or []
        return self._roster_in_memory.get(jid.bare()) or []

    def feed(self, jid: JID, element: ET.Element):
        if len(element) != 1:
            return SE.invalid_xml()

        return self._handlers[element.attrib.get("type")](jid, element)

    def handle_get(self, jid: JID, element: ET.Element):
        if jid.domain == metadata.HOST:
            jid = jid.user
        else:
            jid = jid.bare()

        roster = self._roster_in_memory.get(jid)

        iq = IQ(type_=IQ.TYPE.RESULT, id_=element.attrib.get("id"))
        query = ET.SubElement(iq, "query", attrib={"xmlns": "jabber:iq:roster"})

        for item in roster or []:
            query.append(ET.fromstring(item.get("item")))

        return ET.tostring(iq)

    def handle_set(self, jid: JID, element: ET.Element):
        query = element.find("{jabber:iq:roster}query")
        if jid.domain == metadata.HOST:
            jid = jid.user
        else:
            jid = jid.bare()

        if query is None:
            return SE.invalid_xml()

        new_item = query.findall("{jabber:iq:roster}item")

        if len(new_item) != 1:
            return SE.invalid_xml()

        new_item = new_item[0]
        # remove = None
        # if "subscription" in new_item.attrib.keys():
        #     remove = new_item.attrib["subscription"] == "remove"

        roster = self._roster_in_memory.get(jid)
        if roster:
            match_item = [i for i in roster if ET.fromstring(i.get("item")).get("jid") == new_item.attrib.get("jid")]
            if match_item:
                match_item = match_item[0]
                if new_item.attrib.get("remove") == "remove":
                    with DB.connection() as con:
                        query = delete(Model.Roster).where(
                            Model.Roster.c.jid == jid
                            and Model.Roster.c.roster_item == ET.tostring(match_item.get("item")).decode()
                        )
                        con.execute(query)
                        # con.execute("DELETE FROM roster WHERE jid = ? AND rosterItem = ?",
                        #             (jid, ET.tostring(match_item.get("item")).decode()))
                        con.commit()
                else:
                    with DB.connection() as con:
                        query = update(Model.Roster).where(
                            Model.Roster.c.jid == jid
                            and Model.Roster.c.roster_item == match_item.get("item")
                        ).values({"roster_item": ET.tostring(new_item).decode()})
                        con.execute(query)
                        # con.execute("UPDATE roster SET rosterItem = ? WHERE jid = ? AND rosterItem = ?",
                        #             (ET.tostring(new_item).decode(), jid, match_item.get("item")))
                        con.commit()
            else:
                if new_item.attrib.get("remove") != "remove":
                    with DB.connection() as con:
                        query = insert(Model.Roster).values({
                            "jid": jid,
                            "roster_item": ET.tostring(new_item).decode()
                        })
                        con.execute(query)
                        # con.execute("INSERT INTO roster(jid, rosterItem) VALUES (?, ?)",
                        #             (jid, ET.tostring(new_item).decode()))
                        con.commit()

        else:
            with DB.connection() as con:
                query = insert(Model.Roster).values({
                    "jid": jid,
                    "roster_item": ET.tostring(new_item).decode()
                })
                con.execute(query)
                # con.execute("INSERT INTO roster(jid, rosterItem) VALUES (?, ?)",
                #             (jid, ET.tostring(new_item).decode()))
                con.commit()

        self._update_roster()
        res = IQ(id_=element.attrib.get("id"), type_=IQ.TYPE.RESULT)
        return ET.tostring(res)

    def handle_result(self, _, __):
        # It's safe to ignore this stanza
        return



import json
import os
import pickle
import socket
from dataclasses import dataclass, replace, field

from typing import List, Tuple


@dataclass(frozen=True)
class Parameters:
    host: str = "localhost"
    client_port: int = 5222
    server_port: int = 5269
    server_out_port: int = 5269
    family: socket.AddressFamily = socket.AF_INET
    connection_timeout: int = 60
    database_path: str = os.path.join(os.getcwd(), "pyjabber.db")
    database_purge: bool = False
    database_in_memory: bool = False
    cert_path: str = None
    message_persistence: bool = True
    plugins: List[str] = field(default_factory=lambda: [
            'http://jabber.org/protocol/disco#info',
            'http://jabber.org/protocol/disco#items',
            'http://jabber.org/protocol/pubsub',
            'http://jabber.org/protocol/pubsub#publish',
            'http://jabber.org/protocol/pubsub#subscribe',
            'http://jabber.org/protocol/pubsub#config-node',
            'http://jabber.org/protocol/pubsub#create-nodes',
            'http://jabber.org/protocol/pubsub#delete-nodes',
            'jabber:iq:register',
            'jabber:x:data',
            'urn:xmpp:ping',
            'jabber:iq:rpc',
    ])

    items: List[tuple] = field(default_factory=lambda: [
        ('pubsub', 'service', 'http://jabber.org/protocol/pubsub')
    ])

    def dump(self, file_path: str):
        with open(file_path, "wb") as data:
            pickle.dump(self, data)

    def update_from_json(self, file_path: str):
        with open(file_path, "r") as file:
            loaded = json.load(file)
            return replace(
                self,
                host=loaded.get('host') or "localhost",
                client_port=loaded.get('client_port', 5222),
                server_port=loaded.get('server_port', 5269),
                server_out_port=loaded.get('server_out_port', 5269),
                family=loaded.get('family', socket.AF_INET),
                connection_timeout=loaded.get('connection_timeout', 60),
                database_path=loaded.get('database_path', os.path.join(os.getcwd(), "pyjabber.db")),
                database_purge=loaded.get('database_purge', False),
                database_in_memory=loaded.get('database_in_memory', False),
                cert_path=loaded.get('cert_path', None),
                message_persistence=loaded.get('message_persistence', True),
                plugins=loaded.get('plugins', []),
                items=[tuple(item) for item in loaded.get('items', [])]
            )

    @staticmethod
    def load(file_path: str):
        with open(file_path, "rb") as data:
            return pickle.load(data)

    @staticmethod
    def from_json(file_path: str):
        with open(file_path, "r") as file:
            loaded = json.load(file)
            return Parameters(
                host=loaded.get('host') or "localhost",
                client_port=loaded.get('client_port', 5222),
                server_port=loaded.get('server_port', 5269),
                server_out_port=loaded.get('server_out_port', 5269),
                family=loaded.get('family', socket.AF_INET),
                connection_timeout=loaded.get('connection_timeout', 60),
                database_path=loaded.get('database_path', os.path.join(os.getcwd(), "pyjabber.db")),
                database_purge=loaded.get('database_purge', False),
                database_in_memory=loaded.get('database_in_memory', False),
                cert_path=loaded.get('cert_path', None),
                message_persistence=loaded.get('message_persistence', True),
                plugins=loaded.get('plugins', []),
                items=[tuple(item) for item in loaded.get('items', [])]
            )



import os.path
from typing import List

HOST = None
IP = None
CONFIG_PATH = None
CERT_PATH = None
ROOT_PATH = None
DATABASE_PATH = None
DATABASE_IN_MEMORY = None
DATABASE_PURGE = None
MESSAGE_PERSISTENCE = None
PLUGINS = None
ITEMS =None

TLS_QUEUE = None
CONNECTION_QUEUE = None
MESSAGE_QUEUE = None


def init_config(
    host: str,
    ip: List[str],
    config_path: str,
    cert_path: str,
    root_path:str,
    database_path: str,
    database_in_memory: bool,
    database_purge: bool,
    message_persistence: bool,
    plugins: List[str],
    items: List[tuple]
):
    global HOST, IP, CONFIG_PATH, CERT_PATH, ROOT_PATH, DATABASE_PATH, DATABASE_IN_MEMORY, DATABASE_PURGE, MESSAGE_QUEUE, MESSAGE_PERSISTENCE, PLUGINS, ITEMS
    HOST = host
    IP = ip
    CONFIG_PATH = config_path
    CERT_PATH = cert_path
    ROOT_PATH = root_path
    DATABASE_PATH = database_path
    DATABASE_IN_MEMORY = database_in_memory
    DATABASE_PURGE = database_purge
    MESSAGE_PERSISTENCE = message_persistence
    PLUGINS = plugins
    ITEMS = items

    global TLS_QUEUE, CONNECTION_QUEUE, MESSAGE_QUEUE




from asyncio import Transport
from enum import Enum
from typing import Union
from uuid import uuid4
from xml.etree import ElementTree as ET

from pyjabber.features import InBandRegistration as IBR
from pyjabber.features.StartTLSFeature import StartTLSFeature, proceed_response
from pyjabber.features.StreamFeature import StreamFeature
from pyjabber.features.SASLFeature import SASLFeature, SASL
from pyjabber.features.ResourceBinding import ResourceBinding
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber import metadata
from pyjabber.stanzas.IQ import IQ
from pyjabber.stanzas.error import StanzaError as SE

class Stage(Enum):
    """
    Stream connection states.
    """
    CONNECTED = 0
    OPENED = 1
    SSL = 2
    SASL = 3
    AUTH = 4
    BIND = 5
    READY = 6


class Signal(Enum):
    RESET = 0
    DONE = 1
    FORCE_CLOSE = 2

    def __eq__(self, other):
        if not isinstance(other, Signal):
            return False
        return self.value == other.value


class StreamHandler:
    def __init__(self, transport, starttls) -> None:
        self._host = metadata.HOST
        self._transport = transport
        self._starttls = starttls
        self._streamFeature = StreamFeature()
        self._connection_manager: ConnectionManager = ConnectionManager()
        self._stage = Stage.CONNECTED

        self._stages_handlers = {
            Stage.CONNECTED: self._handle_init,
            Stage.OPENED: self._handle_tls,
            Stage.SSL: self._handle_init_ssl,
            Stage.SASL: self._handle_ssl,
            Stage.AUTH: self._handle_init_resource_bind,
            Stage.BIND: self._handle_resource_bind
        }

    @property
    def transport(self) -> Transport:
        return self._transport

    @transport.setter
    def transport(self, transport: Transport):
        self._transport = transport

    def handle_open_stream(self, elem: ET.Element = None) -> Union[Signal, None]:
        try:
            return self._stages_handlers[self._stage](elem)
        except KeyError:
            SE.internal_server_error()

    def _handle_init(self, _):
        self._streamFeature.reset()
        self._streamFeature.register(StartTLSFeature())
        self._transport.write(self._streamFeature.to_bytes())

        self._stage = Stage.OPENED

    def _handle_tls(self, element: ET.Element):
        if "starttls" in element.tag:
            self._transport.write(proceed_response())
            self._starttls()
            self._stage = Stage.SSL
            self._transport.pause_reading()
            return Signal.RESET

        else:
            raise Exception()

    def _handle_init_ssl(self, _):
        self._streamFeature.reset()

        self._streamFeature.register(IBR.InBandRegistration())
        self._streamFeature.register(SASLFeature())
        self._transport.write(self._streamFeature.to_bytes())

        self._stage = Stage.SASL

    def _handle_ssl(self, element: ET.Element):
        res = SASL().feed(element, {"peername": self._transport.get_extra_info('peername')})
        if type(res) is tuple:
            if res[0].value == Signal.RESET.value:
                self._transport.write(res[1])
                self._stage = Stage.AUTH
                return Signal.RESET
            else:
                self._transport.write(res[1])
                self._stage = Stage.AUTH
                return

        self._transport.write(res)

    def _handle_init_resource_bind(self, _):
        self._streamFeature.reset()
        self._streamFeature.register(ResourceBinding())
        self._transport.write(self._streamFeature.to_bytes())

        self._stage = Stage.BIND

    def _handle_resource_bind(self, element: ET.Element):
        if "iq" in element.tag:
            if element.attrib.get("type") == "set":
                resource_id = str(uuid4())

                iq_res = IQ(type_=IQ.TYPE.RESULT, id_=element.get('id') or str(uuid4()))
                bind_res = ET.SubElement(iq_res, "bind", attrib={"xmlns": "urn:ietf:params:xml:ns:xmpp-bind"})

                peername = self._transport.get_extra_info('peername')
                new_jid = self._connection_manager.get_jid(peername)
                new_jid.resource = resource_id

                ET.SubElement(bind_res, 'jid').text = str(new_jid)

                self._transport.write(ET.tostring(iq_res))

                """
                Stream is negotiated.
                Update the connection register with the jid and transport
                """
                self._connection_manager.set_jid(peername, new_jid, self._transport)

        return Signal.DONE




from xml.etree import ElementTree as ET

from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber import metadata


class Ping:
    @staticmethod
    def feed(jid: JID, element: ET.Element):
        if element.attrib.get('to') == metadata.HOST:
            return ET.tostring(
                IQ(
                    type_=IQ.TYPE.RESULT,
                    id_=element.attrib.get('id'),
                    from_=metadata.HOST,
                    to=str(jid)
                )
            )



import os

import sqlalchemy
from alembic import command
from alembic.config import Config
from sqlalchemy import create_engine, MetaData, Engine, text

from pyjabber import metadata
from loguru import logger

from pyjabber.db.model import Model


class DB:
    _engine: sqlalchemy.Engine = None

    @staticmethod
    def connection() -> sqlalchemy.Connection:  #pragma: no cover
        """
        Returns an already crafted connection with the database.
        It takes the parameters from the server class instance (i.e., DB path | DB in memory)
        """
        return DB._engine.connect()

    @staticmethod
    def close_engine():
        DB._engine.dispose()

    @staticmethod
    def setup_database() -> Engine:
        if metadata.DATABASE_IN_MEMORY:
            logger.info("Using database on memory. ANY CHANGE WILL BE LOST AFTER SERVER SHUTDOWN!")
            DB._engine = create_engine("sqlite:///:memory:")
            DB.init_metadata(DB._engine)

        elif os.path.isfile(metadata.DATABASE_PATH):
            DB._engine = create_engine(f"sqlite:///{metadata.DATABASE_PATH}")
            if metadata.DATABASE_PURGE:
                purge_md = MetaData()
                purge_md.reflect(bind=DB._engine)
                purge_md.drop_all(bind=DB._engine)
                del purge_md
                DB.init_metadata(DB._engine)

        else:
            logger.info("No database found. Initializing one...")
            DB._engine = create_engine(f"sqlite:///{metadata.DATABASE_PATH}")
            DB.init_metadata(DB._engine)

        return DB._engine

    @staticmethod
    def init_metadata(engine: Engine):
        Model.server_metadata.create_all(engine)

    @staticmethod
    def run_db_migrations() -> None:
        cfg = Config()
        cfg.set_main_option("script_location", os.path.join(metadata.ROOT_PATH, '..', 'alembic_local'))
        cfg.set_main_option("sqlalchemy.url", DB.get_database_url_sqlite())
        command.upgrade(cfg, "head")

    @staticmethod
    def get_database_url_sqlite() -> str:
        path = metadata.DATABASE_PATH
        return f"sqlite:///{path}"



import xml.etree.ElementTree as ET
from contextlib import closing
from typing import List
from uuid import uuid4

from sqlalchemy import select

from pyjabber.db.database import DB
from pyjabber.db.model import Model
from pyjabber.plugins.roster.Roster import Roster
from pyjabber.stream.JID import JID


def retrieve_roster(jid: str) -> List[str]:  # pragma: no cover
    with DB.connection() as con:
        query = select(Model.Roster).where(Model.Roster.c.jid == jid)
        roster = con.execute(query)
    return roster


def update(id: int, item: ET) -> str:  # pragma: no cover
    with DB.connection() as con:
        query = Model.Roster.update().where(Model.Roster.c.id == id).values(rosterItem=ET.tostring(item).decode())
        # con.execute("UPDATE roster SET rosterItem = ? WHERE id = ?",
        #             (ET.tostring(item).decode(), id))
        con.execute(query)
        con.commit()
        retrieve = select(Model.Roster.c.rosterItem).where(Model.Roster.c.id == id)
        # res = con.execute("SELECT rosterItem from roster WHERE id = ?", (id,))
        res = con.execute(retrieve).first()

    return res[0] if res else ""


def store_pending_sub(from_: str, to_: str, item: ET.Element) -> None:
    with DB.connection() as con:
        con.execute("INSERT INTO pendingsub values (?, ?, ?)", (from_, to_, ET.tostring(item).decode()))
        con.commit()


def check_pending_sub() -> List[str]:
    with DB.connection() as con:
        res = con.execute("SELECT * FROM pendingsub")
        pending = res.fetchall()

        con.execute("DELETE FROM pendingsub")
        con.commit()
    return pending


def check_pending_sub_to(jid: str, to: str) -> ET.Element:
    with DB.connection() as con:
        res = con.execute("SELECT * FROM pendingsub WHERE jid_from = ? AND jid_to = ?", (jid, to))
        res = res.fetchone()
    if res:
        return res


def create_roster_entry(jid: JID, to: JID, roster_manager: Roster):
    iq = ET.Element(
        "iq", attrib={"from": str(jid), "id": str(uuid4()), "type": "set"}
    )
    query = ET.Element("{jabber:iq:roster}query")
    item = ET.Element(
        "{jabber:iq:roster}item",
        attrib={
            "jid": str(to),
            "subscription": "none"})
    query.append(item)
    iq.append(query)

    return roster_manager.feed(jid, iq)




import socket

from loguru import logger


def setup_query_local_ip():
    """Return the local IP of the host machine"""
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.254.254.254', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


def setup_ip_by_host(host: str):
    try:
        return socket.gethostbyname(host)
    except socket.gaierror as e:
        logger.error(e)
        return None


def get_migrations_path() -> str:
    # Localiza la carpeta migrations dentro del paquete instalado
    return str(pkg_resources.files("my_package").joinpath("migrations"))


def run_db_migrations() -> None:
    # 1) Crea la Config de Alembic “en memoria”
    cfg = Config()
    cfg.set_main_option("script_location", get_migrations_path())
    cfg.set_main_option("sqlalchemy.url", os.getenv("DATABASE_URL"))
    # 2) Sube la base de datos hasta la última revisión
    command.upgrade(cfg, "head")



import asyncio
import os
import signal

from loguru import logger

from pyjabber import init_utils
from pyjabber.db.database import DB
from pyjabber.network.XMLProtocol import XMLProtocol
from pyjabber.network.server.incoming.XMLServerIncomingProtocol import XMLServerIncomingProtocol
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.server_parameters import Parameters
from pyjabber.webpage.adminPage import AdminPage
from pyjabber import metadata
from pyjabber.workers import tls_worker, queue_worker

SERVER_FILE_PATH = os.path.dirname(os.path.abspath(__file__))


class Server:
    """Server class

    :param param: Instance of Parameter class, with all the configuration available for the server. If not provided,
    a default profile will be loaded
    """

    def __init__(self, param: Parameters = Parameters()):
        # Server
        self._host = param.host
        self._public_ip = init_utils.setup_query_local_ip()
        self._host_ip = init_utils.setup_ip_by_host(self._host)
        self._client_port = param.client_port
        self._server_port = param.server_port
        self._server_out_port = param.server_out_port
        self._family = param.family
        self._client_listener = None
        self._server_listener = None
        self._adminServer = AdminPage()
        self._connection_timeout = param.connection_timeout

        # Database
        self._database_path = param.database_path
        self._database_purge = param.database_purge
        self._database_in_memory = param.database_in_memory

        # Certs
        self._cert_path = param.cert_path or os.path.join(SERVER_FILE_PATH, "network", "certs")

        # Singletons
        self._connection_manager = ConnectionManager()

        # Global constants to use across the server modules/classes
        metadata.init_config(
            host=param.host,
            ip=[self._host_ip, self._public_ip],
            database_path=self._database_path,
            database_purge=self._database_purge,
            database_in_memory=self._database_in_memory,
            config_path=os.path.join(os.path.dirname(os.path.abspath(__file__)), "config/config.yaml"),
            cert_path=self._cert_path,
            root_path=SERVER_FILE_PATH,
            message_persistence=param.message_persistence or False,
            plugins=param.plugins,
            items=param.items
        )

        # Flags
        self._ready = asyncio.Event()

    @property
    def ready(self):
        return self._ready

    async def run_server(self):
        """
        Launches the configured server, and returns a coroutine.
        """
        try:
            logger.info("Starting server...")

            engine = DB.setup_database()
            if not self._database_in_memory:
                DB.run_db_migrations()
            # if not self._database_in_memory and DB.needs_upgrade(engine):
            #     DB.run_migrations_if_needed()
            self._public_ip = init_utils.setup_query_local_ip()

            loop = asyncio.get_running_loop()

            try:
                self._client_listener = await loop.create_server(
                    lambda: XMLProtocol(
                        namespace="jabber:client",
                        host=self._host,
                        connection_timeout=self._connection_timeout,
                        cert_path=self._cert_path,
                    ),
                    host=[self._host, self._public_ip] if self._public_ip else [self._host],
                    port=self._client_port,
                    family=self._family
                )
            except OSError as e:
                logger.error(e)
                self.raise_exit()

            logger.info(f"Client domain => {self._host}")
            logger.info(
                f"Server is listening clients on {[s.getsockname() for s in self._client_listener.sockets if s]}")

            try:
                self._server_listener = await loop.create_server(
                    lambda: XMLServerIncomingProtocol(
                        namespace="jabber:server",
                        host=self._host,
                        connection_timeout=self._connection_timeout,
                        cert_path=self._cert_path,
                    ),
                    host=["0.0.0.0"],
                    port=self._server_port,
                    family=self._family
                )
            except OSError as e:
                logger.error(e)
                raise SystemExit

            logger.info(
                f"Server is listening servers on {[s.getsockname() for s in self._server_listener.sockets if s]}")
            logger.success("Server started...")
            self._ready.set()

            while True:
                # Keep the coroutine alive, in order to catch the CancelledError exception
                await asyncio.sleep(3600)

        except asyncio.CancelledError:
            logger.info("Stopping server...")

            DB.close_engine()

            if self._client_listener and self._client_listener.is_serving():
                self._client_listener.close()
                await self._client_listener.wait_closed()

            if self._server_listener and self._server_listener.is_serving():
                self._server_listener.close()
                await self._server_listener.wait_closed()

            logger.success("Server stopped...")

    def raise_exit(self, *args):
        """Exception used to signal the safe close of the server"""
        raise SystemExit(1)

    async def start(self):
        """Start the already created and configuration server"""
        metadata.TLS_QUEUE = asyncio.Queue()
        metadata.CONNECTION_QUEUE = asyncio.Queue()
        metadata.MESSAGE_QUEUE = asyncio.Queue()

        signal.signal(signal.SIGINT, self.raise_exit)
        signal.signal(signal.SIGABRT, self.raise_exit)
        signal.signal(signal.SIGTERM, self.raise_exit)

        tasks = None

        try:
            tasks = [
                asyncio.create_task(self._adminServer.start()),
                asyncio.create_task(tls_worker()),
                asyncio.create_task(queue_worker()),
                asyncio.create_task(self.run_server())
            ]
            await asyncio.gather(*tasks)

        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError) as e:  # pragma: no cover
            logger.trace(f"Signal {e.__class__.__name__} intercepted. Stopping server")

        finally:
            for task in tasks:
                _ = task.cancel()

            await asyncio.gather(*tasks)
